/*
		Animation.m


		This class allows for animation.
		
		Created by Ian Howle on 6/7/2005.
		Public Domain.
*/




#import "Animation.h"



@implementation Animation



- (id)init
{
	if ( self = [super init] )
	{
		AnimTimer = nil;
		FrameFlipTarget = nil;
		FrameFlipActionMethod = nil;
		
		FrameCount = 0;
		StartFrame = 0;
		EndFrame = 0;
		AnimLoopCount = 0;
		AnimDir = ANIM_PLAY_FORWARD;
		IndexOfCurrentFrame = -1;
	}
	
	
	return	self;
}





- (void)dealloc
{
	if ( AnimTimer != nil )
	{
		[AnimTimer invalidate];
	}

	[super dealloc];
}





/*****
		setTargetForFrameFlip:action:
		
		This method sets the target and action methods for the
		object that is to receive frame flipping information.
*****/

- (void)setTargetForFrameFlip:(id)target action:(SEL)action
{
	FrameFlipTarget = target;
	FrameFlipActionMethod = action;
}





/*****
		setTargetForAnimDone:action:
		
		This method sets the target and action methods for the object
		that is to be notified when an animation has finished.
*****/

- (void)setTargetForAnimDone:(id)target action:(SEL)action
{
	AnimDoneTarget = target;
	AnimDoneActionMethod = action;
}





/*****
		setAnimLoopCount:
		
		This method sets how the animation will be shown. A value less than one will
		animate indefinitely. A value greater than zero will loop that many times.
*****/

- (void)setAnimLoopCount:(int)animCount
{
	AnimLoopCount = animCount;
}





/*****
		setAnimDir:
		
		This method determines which way the animation will play.
*****/

- (void)setAnimDir:(int)dir
{
	AnimDir = dir;
}





/*****
		setNumOfFrames:
		
		This method sets the number of frames for this animation.
*****/

- (void)setNumOfFrames:(int)frameCount
{
	FrameCount = frameCount;
	
	StartFrame = 0;
	EndFrame = frameCount - 1;
}





/*****
		setFrameRangeFrom:to:
		
		This method allows for playing only part of an animation; from the passed-in
		start frame to the passed-in stop frame.
*****/

- (void)setFrameRangeFrom:(int)fromFrame to:(int)toFrame
{
	if ( fromFrame >= FrameCount )
	{
		fromFrame = FrameCount - 1;
	}
	
	if ( fromFrame < 0 )
	{
		fromFrame = 0;
	}
	
	
	if ( toFrame >= FrameCount )
	{
		toFrame = FrameCount - 1;
	}

	if ( toFrame < 0 )
	{
		toFrame = 0;
	}
	
	
	StartFrame = fromFrame;
	EndFrame = toFrame;
	
	
	if ( StartFrame < EndFrame )
	{
		IndexOfCurrentFrame = StartFrame - 1;
	}
	else
	{
		IndexOfCurrentFrame = StartFrame + 1;
	}
}





/*****
		startAnimWithDelayBetweenFrames:
		
		This method starts the animation.
*****/

- (void)startAnimWithDelayBetweenFrames:(float)timeDelay
{
	if ( AnimTimer != nil )
	{
		[AnimTimer invalidate];
	}

	
	AnimTimer = [NSTimer scheduledTimerWithTimeInterval:timeDelay target:self selector:@selector(animate:) userInfo:nil repeats:YES];
}





/*****
		animate:
		
		This method animates each time it's called by the animation timer.
*****/

- (void)animate:(NSTimer *)animTime
{
	int				tempStartFrame;
	BOOL			endOfLoopFlag = NO;
	
	
	
	// Check if an animation loop has finished.
	
	if ( StartFrame > EndFrame )		// Is the animation going from lowest frame number to highest?
	{
		if ( --IndexOfCurrentFrame < EndFrame )		// Is this the last frame?
		{
			endOfLoopFlag = YES;
		}
	}
	else	// Animation is going from highest frame to lowest.
	{
		if ( ++IndexOfCurrentFrame > EndFrame )		// Is this the last frame?
		{
			endOfLoopFlag = YES;
		}
	}
	
	
	
	if ( endOfLoopFlag )		// Has the animation loop ended yet?
	{
		if ( AnimLoopCount > 0 )		// Any more animation loops? 
		{
			if ( --AnimLoopCount <= 0 )
			{
				[self stopAnim];
				
				return;
			}
		}
		
		
		if ( AnimDir == ANIM_PLAY_FORWARD_AND_BACKWARD )
		{
			tempStartFrame = StartFrame;

			StartFrame = EndFrame;
			EndFrame = tempStartFrame;
			
			if ( StartFrame < EndFrame )
			{
				IndexOfCurrentFrame = StartFrame + 1;
			}
			else
			{
				IndexOfCurrentFrame = StartFrame - 1;
			}
		}
		else
		{
			IndexOfCurrentFrame = StartFrame;
		}
	}
	

	
	// Send a message indicating that a frame has been flipped.

	if ( FrameFlipTarget != nil )
	{
		[FrameFlipTarget performSelector:FrameFlipActionMethod withObject:self];
	}
}





/*****
		stopAnim

		This method stops the animation.
*****/

- (void)stopAnim
{
	if ( AnimTimer != nil )
	{
		[AnimTimer invalidate];
		AnimTimer = nil;
		
		
		
		// Send a message indicating that the animation has stopped. If there is no target and action
		// to call, then a notification is sent.

		if ( AnimDoneTarget != nil )
		{
			[AnimDoneTarget performSelector:AnimDoneActionMethod withObject:self];
		}
		else
		{
			[[NSNotificationCenter defaultCenter] postNotificationName:@"animation done" object:self];
		}
	}
}





/*****
		finishAnim
		
		This method stops an animation after it finishes it's current loop. This is useful
		for gracefully stopping an indefinite animation loop.
*****/

- (void)finishAnim
{
	AnimLoopCount = 1;
}





/*****
		yes/no = isAnimating

		This method returns YES if animation is taking place, otherwise NO is returned.
*****/

- (BOOL)isAnimating
{
	if ( AnimTimer != nil )
	{
		return	[AnimTimer isValid];
	}
	
	
	return	NO;
}





/*****
		index of current frame = currentFrame
		
		This method returns the index of the current frame.
*****/

- (int)currentFrame
{
	return	IndexOfCurrentFrame;
}


		
@end
