/*
		GameSprite.h

	
		Header file for GameSprite.m

		Created by Ian Howle on 2/19/2005.
		Public Domain.
*/




#import "HWNSImage.h"
#import "Animation.h"
#import "GameUtilities.h"



enum	// Indicators of whether a sprite has reached it's boundary or not.
{
	SPRITE_REACHED_NO_BOUNDARY,
	SPRITE_REACHED_LEFT_BOUNDARY,
	SPRITE_REACHED_RIGHT_BOUNDARY,
	SPRITE_REACHED_TOP_BOUNDARY,
	SPRITE_REACHED_BOTTOM_BOUNDARY
};



@interface GameSprite : NSObject <NSCopying>
{
	id					CollisionTarget;		// Sprite collision target.
	SEL					CollisionAction;		// Refers method a called when this sprite has collided.
	
	id					ReachedBoundsTarget;	// Sprite bounds reached target.
	SEL					ReachedBoundsAction;	// Refers to a method called when this sprite has reached or exceeded it's bounds.

	Animation			*AnimObj;				// A reference to this sprite's animation object.
		
	NSString			*Identifier;			// The sprite's unique identifier.
	NSImage				*MasterImage;			// This image holds all frames of animation used by this sprite.
	NSImage				*OriginalSpriteImage;	// When the sprite image is loaded, it gets stored here and is only altered by horizontal and vertical flips.
	NSImage				*SpriteImage;			// The sprite's current image.
	NSArray				*SpriteFrames;			// A list of frames used by this sprite for animation.
	NSShadow			*SpriteShadow;			// A shadow for this sprite.
	NSShadow			*NoShadow;				// If this sprite has a shadow, then NoShadow will unset it so other drawing afterwards won't get this shadow.
	NSTimer				*StrobeTimer;			// This timer controls the sprite's strobing effect.
	NSTimer				*FadeTimer;				// This timer controls the time between changes in opacity levels.
	NSTimer				*SpinTimer;				// This timer controls how fast this sprite will spin.
	NSTimer				*ScaleTimer;			// This timer controls how fast this sprite will scale to it's desired size.

	NSRect				PreviousFrame;			// The sprite frame since the last time the draw method was called.
	NSRect				DirtyRect;				// This covers the size of the sprite plus any extra space needed to clean any stray pixels.
	NSSize				DesiredSize;			// The sprite's desired size.
	NSPoint				SpriteDir;				// Direction sprite will travel in.
	NSPoint				SpritePos;				// The sprite's current position.
	NSPoint				SpriteOffsetPos;		// This is used for offsetting the sprite when it's rectangle size changes (e.g. during rotation).

	float				Opacity;				// The sprite's current opacity level.
	float				SavedOpacity;			// This is where the sprite's normal opacity level is stored.
	float				FadeFromOpacity;		// The opacity level to begin fading from.
	float				FadeToOpacity;			// The destination opacity level to fade to.
	float				Rotation;				// The sprite's current angle of rotation.
	float				DegreesToRotate;		// Number of degree to rotate (this determines how smooth the sprite will spin).

	float				LeftBoundary;			// These are the boundaries that the sprite cannot pass.
	float				TopBoundary;			// ...
	float				RightBoundary;			// ...
	float				BottomBoundary;			// ...

	float				CollisionLeftOffset;	// The distance from the left edge of the sprite frame outward.
	float				CollisionRightOffset;	// ...
	float				CollisionTopOffset;		// ...
	float				CollisionBottomOffset;	// ...

	BOOL				ActiveFlag;					// This flag determines whether this sprite is active or not.
	BOOL				CollisionDetectionFlag;		// This flag marks whether collision detection is on or off.
	BOOL				AutoConfineToBoundaryFlag;	// This flag indicates whether this sprite should automatically be confined to it's defined boundary.
	BOOL				VisibleFlag;				// This flag marks the sprite as being either visible or invisible.
}



// Instance methods...

- (id)initWithIndentifier:(NSString *)name;

- (void)setupSpriteWithMasterImage:(NSImage *)mImage withFrames:(NSArray *)frames atPos:(NSPoint)pos inDir:(NSPoint)dir withOpacity:(float)sOpacity;
- (void)setupSpriteWithImage:(NSImage *)sImage atPos:(NSPoint)pos inDir:(NSPoint)dir withOpacity:(float)sOpacity;
- (void)replaceImageAtFrameIndex:(int)frameIndex withImage:(NSImage *)image;

- (void)setCollisionTarget:(id)target action:(SEL)action;
- (void)setReachedBoundsTarget:(id)target action:(SEL)action;

- (NSString *)identifier;
- (NSImage *)spriteImage;
- (NSImage *)imageAtFrameIndex:(int)frameIndex;
- (void)setActive:(BOOL)active;
- (BOOL)isActive;
- (void)setCanCollide:(BOOL)canCollide;
- (BOOL)canCollide;
- (void)setVisible:(BOOL)visible;
- (BOOL)isVisible;
- (void)setOpacity:(float)level;
- (float)opacity;
- (NSSize)size;
- (NSRect)bounds;
- (NSRect)frame;
- (float)rightFrameEdge;
- (float)topFrameEdge;

- (void)setPosX:(float)x posY:(float)y;
- (void)setPosX:(float)x;
- (void)setPosY:(float)y;
- (void)setOrigin:(NSPoint)p;
- (NSPoint)origin;

- (void)setDir:(NSPoint)dir;
- (void)setDirX:(float)x dirY:(float)y;
- (void)setDirX:(float)x;
- (void)setDirY:(float)y;
- (NSPoint)spriteDir;

- (void)setSpeed:(float)speed;
- (void)setSpeedX:(float)xSpeed speedY:(float)ySpeed;

- (void)move;

- (Animation *)animRef;
- (void)removeAnimObject;
- (void)nextFrame:(id)sender;
- (void)setFrame:(int)nextFrame;

- (void)newWidth:(float)width newHeight:(float)height center:(BOOL)center;
- (void)scaleByPercent:(float)percent center:(BOOL)center;
- (void)scaleToPercentage:(float)scaleTo byPixelInc:(float)pixelInc atRate:(float)rate center:(BOOL)center;
- (void)scaleToOriginalSizeByPixelInc:(float)pixelInc atRate:(float)rate center:(BOOL)center;
- (void)scale:(NSTimer *)timer;

- (void)rotateByDegrees:(float)deg;
- (void)setSpinRate:(float)rate degreesPerRotation:(float)degPerRotation;
- (float)spinRate;
- (void)spin:(NSTimer *)timer;

- (void)shearXBy:(float)xShear yBy:(float)yShear;

- (void)strobeAtRate:(float)rate;
- (float)strobeRate;
- (void)strobe:(NSTimer *)timer;

- (void)fadeTo:(float)level atRate:(float)rate;
- (void)fadeFrom:(float)from fadeTo:(float)to atRate:(float)rate;
- (void)fadeFrom:(float)from fadeTo:(float)to inIncrementsOf:(float)fadeIncrement atRate:(float)rate;
- (void)fading:(NSTimer *)timer;
- (void)stopFading;

- (void)hMirror;
- (void)vMirror;
- (void)flipSpriteHorizontally:(BOOL)horizontal;

- (void)setShadowOffset:(NSSize)offset color:(NSColor *)color blur:(float)blur;
- (NSShadow *)shadow;
- (void)removeShadow;

- (void)setDirtyRect:(NSRect)dRect;
- (NSRect)dirtyRect;
- (void)draw;

- (void)setBoundaryLeft:(float)left top:(float)top right:(float)right bottom:(float)bottom autoConfineToBoundary:(BOOL)confine;
- (void)setBoundaryRect:(NSRect)bRect autoConfineToBoundary:(BOOL)confine;
- (NSRect)boundaryRect;
- (void)setAutoConfineToBoundary:(BOOL)confine;
- (BOOL)autoConfinedToBoundary;
- (int)boundaryReached;
- (int)checkAndConfineToBoundary;

- (void)setCollisionAreaLeftOffset:(float)leftOffset topOffset:(float)topOffset rightOffset:(float)rightOffset bottomOffset:(float)bottomOffset;
- (void)setCollisionAreaWidthOffset:(float)widthOffset heightOffset:(float)heightOffset;
- (void)setCollisionAreaOffset:(float)areaOffset;
- (NSRect)collisionArea;
- (BOOL)collidedWithPoint:(NSPoint)thePoint;
- (BOOL)collidedWithRect:(NSRect)theRect;
- (BOOL)collidedWithSprite:(GameSprite *)otherSprite;



@end
