/*
		GameSprite.m


		This class defines a sprite, it's appearance, and behaviour.
		
		Created by Ian Howle on 2/19/2005.
		Public Domain.
*/




#import "GameSprite.h"



@implementation GameSprite



/*****
		self = initWithIndentifier:
		
		This method is called when an instance of this class is created.
*****/

- (id)initWithIndentifier:(NSString *)name
{
	if ( self = [super init] )
	{
		Identifier = [[NSString alloc] initWithString:name];
		
		CollisionTarget = nil;
		ReachedBoundsTarget = nil;
		
		AnimObj = nil;
		
		MasterImage = nil;
		SpriteFrames = nil;
		SpriteImage = nil;
		OriginalSpriteImage = nil;
		
		SpriteShadow = nil;
		NoShadow = nil;
		
		StrobeTimer = nil;
		FadeTimer = nil;
		SpinTimer = nil;
		ScaleTimer = nil;
		
		Opacity = 1.0;
		SavedOpacity = Opacity;
		
		LeftBoundary = 0.0;
		TopBoundary = 0.0;
		RightBoundary = 0.0;
		BottomBoundary = 0.0;
		
		CollisionLeftOffset = 0.0;
		CollisionRightOffset = 0.0;
		CollisionTopOffset = 0.0;
		CollisionBottomOffset = 0.0;
		
		PreviousFrame = NSZeroRect;
		DirtyRect = NSZeroRect;
		DesiredSize = NSZeroSize;
		SpriteOffsetPos = NSZeroPoint;

		ActiveFlag = YES;					// This sprite is active.
		CollisionDetectionFlag = YES;		// Collision detection on.
		VisibleFlag = YES;					// It's visible.
	}


	return	self;
}





/*****
		dealloc
		
		This method is called when an instance of this class is deallocated.
*****/

- (void)dealloc
{
	[Identifier release];		// Free this sprite's identifier.
	[self removeAnimObject];	// Remove the animation object, if it exists.

	[MasterImage release];		// If this sprite has a master image, then free it.

	[SpriteFrames release];		// If this sprite has a list of frames, then free them.
	
	[OriginalSpriteImage release];		// If this sprite has an image, then free it.
	[SpriteImage release];				// Free the copy of the sprite image, if it exists.

	[SpriteShadow release];
	[NoShadow release];
	
	
	if ( StrobeTimer != nil )	// Free the strobe timer?
	{
		[StrobeTimer invalidate];	// If the strobe timer is firing, then stop it.
	}
	
	
	if ( FadeTimer != nil )		// Free the fade timer?
	{
		[FadeTimer invalidate];		// If the fade timer is firing, then stop it.
	}
	
	
	if ( SpinTimer != nil )		// Free the spin timer?
	{
		[SpinTimer invalidate];		// If the spin timer is firing, then stop it.
	}
	
	
	if ( ScaleTimer != nil )	// Free the scale timer?
	{
		[ScaleTimer invalidate];	// If the scale timer is firing, then stop it.
	}
	
	
	[super dealloc];
}





/*****
		self = copyWithZone:
		
		This method creates and returns a copy of this sprite.
		
		NOTE: it is the responsibility of whoever gets this copied
		sprite to release it.
*****/

- (id)copyWithZone:(NSZone *)zone
{
	NSString		*newIdentifier = [NSString stringWithFormat:@"%@ copy", Identifier];
	GameSprite		*copyOfSprite = [[GameSprite alloc] initWithIndentifier:newIdentifier];
	
	
	[copyOfSprite setupSpriteWithMasterImage:MasterImage withFrames:SpriteFrames atPos:SpritePos inDir:SpriteDir withOpacity:Opacity];
	
	
	return	copyOfSprite;
}



#pragma mark -
#pragma mark Sprite Setup



/*****
		setupSpriteWithMasterImage:withFrames:atPos:inDir:withOpacity:
		
		This method does basic initialization for this sprite. It's purpose is to combine
		several initialization methods into one, for convenience.
*****/

- (void)setupSpriteWithMasterImage:(NSImage *)mImage withFrames:(NSArray *)frames atPos:(NSPoint)pos inDir:(NSPoint)dir withOpacity:(float)sOpacity
{
	[MasterImage release];		// If there already is a master image for this sprite, then release it.
	MasterImage = nil;

	
	
	if ( mImage != nil )		// Make sure the passed in master image is valid.
	{
		MasterImage = [mImage retain];
		
		[MasterImage setScalesWhenResized:YES];
	}
	else
	{
		NSLog(@"error- no master image");
		
		return;
	}


	[SpriteFrames release];		// If there already is a list of sprite frames, then release them.
	SpriteFrames = nil;
	
	
	if ( frames != nil && [frames count] > 0 )		// Make sure the passed-in frames array exists.
	{
		int			numOfFrames = [frames count];


		if ( numOfFrames > 1 )		// If there is more than one frame, then set the animation object.
		{
			[[self animRef] setNumOfFrames:numOfFrames];
		}
		
		
		SpriteFrames = [frames retain];
	}
	else
	{
		NSLog(@"error - no sprite frames list");
		
		return;
	}
	
	
	
	[self setFrame:0];					// Start with the first frame of animation.
	[self setPosX:pos.x posY:pos.y];	// Set this sprite's onscreen position.
	[self setDirX:dir.x dirY:dir.y];	// Set direction for this sprite to move in.
	[self setOpacity:sOpacity];			// Set this sprite's opacity.
}





/*****
		setupSpriteWithImage:atPos:inDir:withOpacity:
		
		This is a convenience method that initializes a sprite that has only one frame.
*****/

- (void)setupSpriteWithImage:(NSImage *)sImage atPos:(NSPoint)pos inDir:(NSPoint)dir withOpacity:(float)sOpacity
{
	NSArray				*frameArray = [NSArray arrayWithObject:[NSValue valueWithRect:NSMakeRect( 0.0, 0.0, [sImage size].width, [sImage size].height )]];
	
	
	[self setupSpriteWithMasterImage:sImage withFrames:frameArray atPos:pos inDir:dir withOpacity:sOpacity];
}



#pragma mark -
#pragma mark Target/Actions



/*****
		setCollisionTarget:action:
		
		This method sets the target and action for when this sprite collides.
*****/

- (void)setCollisionTarget:(id)target action:(SEL)action
{
	CollisionTarget = target;
	CollisionAction = action;
}





/*****
		setReachedBoundsTarget:action:
		
		This method sets the target and action for when this sprite reaches
		or exceeds it's boundary.
*****/

- (void)setReachedBoundsTarget:(id)target action:(SEL)action
{
	ReachedBoundsTarget = target;
	ReachedBoundsAction = action;
}



#pragma mark -
#pragma mark Accessors



/*****
		identifier name = identifier
		
		This method returns the sprite's identifier name.
*****/

- (NSString *)identifier
{
	return	Identifier;
}





/*****
		sprite image = spriteImage
		
		This method returns the sprite's current image.
*****/

- (NSImage *)spriteImage
{
	return	SpriteImage;
}





/*****
		setActive:
		
		This method sets whether this sprite is active or not. If it's set to be
		non-active, then any rotation, fading, etc... will stop.
*****/

- (void)setActive:(BOOL)active
{
	ActiveFlag = active;
	
	
	if ( !ActiveFlag )
	{
		if ( StrobeTimer != nil )	// Free the strobe timer?
		{
			[StrobeTimer invalidate];	// If the strobe timer is firing, then stop it.
			StrobeTimer = nil;
		}
		
		
		if ( FadeTimer != nil )		// Free the fade timer?
		{
			[FadeTimer invalidate];		// If the fade timer is firing, then stop it.
			FadeTimer = nil;
		}
		
		
		if ( SpinTimer != nil )		// Free the spin timer?
		{
			[SpinTimer invalidate];		// If the spin timer is firing, then stop it.
			SpinTimer = nil;
		}
		
		
		if ( ScaleTimer != nil )	// Free the scale timer?
		{
			[ScaleTimer invalidate];	// If the scale timer is firing, then stop it.
			ScaleTimer = nil;
		}
	}
	
	
	[self setCanCollide:ActiveFlag];	// An inactive sprite does not have collision detection.
}





/*****
		yes/no = isActive
		
		This method returns YES if this sprite is active. Otherwise, NO is returned.
*****/

- (BOOL)isActive
{
	return	ActiveFlag;
}





/*****
		setCanCollide:
		
		This method sets whether collision detection is on or off.
*****/

- (void)setCanCollide:(BOOL)canCollide
{
	CollisionDetectionFlag = canCollide;
}





/*****
		yes/no = canCollide
		
		This method returns a boolean indicating whether collision detection
		is on or off.
*****/

- (BOOL)canCollide
{
	return	CollisionDetectionFlag;
}





/*****
		setVisible:
		
		This method makes the sprite visible if YES is passed in, or invisible
		if NO is passed in.
*****/

- (void)setVisible:(BOOL)visible
{
	VisibleFlag = visible;
	
	
	if ( VisibleFlag )		// Make sprite visible or invisible?
	{
		[self setOpacity:SavedOpacity];		// Return the sprite to it's designated opacity level.
	}
	else
	{
		SavedOpacity = Opacity;		// Save the current opacity level.
		
		[self setOpacity:0.0];		// Make the sprite invisible.
	}
}





/*****
		yes/no = isVisible
		
		This method returns YES if this sprite is visible and NO if it's invisible.
*****/

- (BOOL)isVisible
{
	return	VisibleFlag;
}





/*****
		setOpacity:
		
		This method sets the sprite's opacity.
*****/

- (void)setOpacity:(float)level
{
	Opacity = level;
}





/*****
		opacity level = opacity
		
		This method returns the sprite's opacity level.
*****/

- (float)opacity
{
	return	Opacity;
}





/*****
		image size = size
		
		This mewthod returns the size of this sprite.
*****/

- (NSSize)size
{
	return	[SpriteImage size];
}





/*****
		sprite position = origin
		
		This method returns the sprite's current position within it's view.
*****/

- (NSPoint)origin
{
	return	SpritePos;
}





/*****
		sprite bounds rectangle = bounds
		
		This method returns the sprite's dimensions.
*****/

- (NSRect)bounds
{
	return	NSMakeRect( 0.0, 0.0, [SpriteImage size].width, [SpriteImage size].height );
}





/*****
		sprite frame = frame
		
		This method returns the sprite's frame within it's view.
*****/

- (NSRect)frame
{
	return	NSMakeRect( SpritePos.x, SpritePos.y, [SpriteImage size].width, [SpriteImage size].height );
}



#pragma mark -
#pragma mark Movement



/*****
		setPosX:posY:
		
		This method positions the sprite within the view.
*****/

- (void)setPosX:(float)x posY:(float)y
{
	SpritePos.x = x;		// Set new x position.
	SpritePos.y = y;		// Set new y position.
}





/*****
		setOrigin:
		
		This method positions the sprite within the view.
*****/

- (void)setOrigin:(NSPoint)p
{
	SpritePos.x = p.x;	// Set new position.
	SpritePos.y = p.y;
}





/*****
		setDir:

		This method sets the direction that the sprite will be moving in.
		A direction of (0.0, 0.0) will prevent the sprite from moving.
*****/

- (void)setDir:(NSPoint)dir
{
	SpriteDir.x = dir.x;
	SpriteDir.y = dir.y;
}





/*****
		setDirX:dirY:
		
		This method sets the direction the sprite will be moving in. An x and y value
		of zero will prevent the sprite from moving.
*****/

- (void)setDirX:(float)x dirY:(float)y
{
	SpriteDir.x = x;		// Set new horizontal direction.
	SpriteDir.y = y;		// Set new vertical direction.
}





/*****
		setDirX:
		
		This method sets the horizontal direction that this sprite will be moving in.
*****/

- (void)setDirX:(float)x
{
	SpriteDir.x = x;
}





/*****
		setDirY:
		
		This method sets the horizontal direction that this sprite will be moving in.
*****/

- (void)setDirY:(float)y
{
	SpriteDir.y = y;
}





/*****
		sprite direction = spriteDir
		
		This method returns this sprite's horizontal and vertical direction.
*****/

- (NSPoint)spriteDir
{
	return	SpriteDir;
}





/*****
		setSpeed:
		
		This method sets both the horizontal and vertical speed
		for this sprite to the passed-in value.

		NOTE: a positive value should be passed-in. This method will
		ensure that this sprite continues in the same direction.
*****/

- (void)setSpeed:(float)speed
{
	[self setSpeedX:speed speedY:speed];
}





/*****
		setSpeedX:speedY:
		
		This method sets the horizontal speed and vertical speed
		for this sprite based on the passed-in values.

		NOTE: a positive value should be passed-in. This method will
		ensure that this sprite continues in the same direction.
*****/

- (void)setSpeedX:(float)xSpeed speedY:(float)ySpeed
{
	if ( SpriteDir.x < 0.0 )		// Set horizontal speed.
	{
		SpriteDir.x = -xSpeed;
	}
	else
	{
		SpriteDir.x = xSpeed;
	}
	
	
	if ( SpriteDir.y < 0.0 )		// Set vertical speed.
	{
		SpriteDir.y = -ySpeed;
	}
	else
	{
		SpriteDir.y = ySpeed;
	}
}





/*****
		move
		
		This metod moves the sprite.
*****/

- (void)move
{
	if ( ActiveFlag )
	{
		SpritePos.x += SpriteDir.x;			// Move sprite in horizontal direction.
		SpritePos.y += SpriteDir.y;			// Move sprite in vertical direction.
		

		
		// Is this sprite automatically confined to it's bounds?
		
		if ( AutoConfineToBoundaryFlag )
		{
			[self checkAndConfineToBoundary];
		}

		[self setDirtyRect:NSUnionRect( PreviousFrame, [self frame] )];		// Set the dirty rectangle.
	}
}



#pragma mark -
#pragma mark Animation



/*****
		animation object reference = animRef
		
		This method returns a reference to this sprite's animation object. If the animation object does not
		exist, then it will be created.
*****/

- (Animation *)animRef
{
	if ( AnimObj == nil )		// Create the animation object if it doesn't already exist.
	{
		AnimObj = [[Animation alloc] init];

		[AnimObj setTargetForFrameFlip:self action:@selector(nextFrame:)];
	}
	
	
	return	AnimObj;
}





/*****
		removeAnimObject
		
		This method releases the animation object (if one exists) for this sprite.
*****/

- (void)removeAnimObject
{
	if ( AnimObj != nil )
	{
		[AnimObj stopAnim];
		[AnimObj release];
		AnimObj = nil;
	}
}





/*****
		nextFrame:
		
		This method is called by the animation object each time it
		flips to another frame.
*****/

- (void)nextFrame:(id)sender
{
	[self setFrame:[sender currentFrame]];
}




		
/*****
		setFrame:
		
		This method sets the next frame of animation.
*****/

- (void)setFrame:(int)nextFrame
{
	NSRect			spriteRect = [[SpriteFrames objectAtIndex:nextFrame] rectValue];	// Get the next frame's rectangle.
	NSSize			spriteSize = spriteRect.size;										// Get the size of the next frame.



	// Create the template image, if it has not already been created.

	[OriginalSpriteImage release];
	OriginalSpriteImage = [[NSImage alloc] initWithSize:spriteSize];



	// Copy the next frame from the master image onto the template image.
	
	[OriginalSpriteImage lockFocus];
	[MasterImage drawInRect:NSMakeRect( 0.0, 0.0, [OriginalSpriteImage size].width, [OriginalSpriteImage size].height ) fromRect:spriteRect operation:NSCompositeSourceOver fraction:1.0];
	[OriginalSpriteImage unlockFocus];

	
	
	// Copy the sprite image data into the template image.

	[OriginalSpriteImage setScalesWhenResized:YES];		// Allow it to be resized. (The template itself should never be resized, but any copies of it will have this ability.)
	
	SpriteImage = [OriginalSpriteImage copy];		// Get a copy of the original.
	DesiredSize = [SpriteImage size];				// Resize the original to the desired size.
}



#pragma mark -
#pragma mark Scaling



/*****
		newWidth:newHeight:
		
		This method resizes this sprite with the desired width and height.
*****/

- (void)newWidth:(float)width newHeight:(float)height center:(BOOL)center
{
	float			oldWidth = [self size].width;
	float			oldHeight = [self size].height;



	DesiredSize = NSMakeSize( width, height );		// Set the desired size for this sprite.
	
	[SpriteImage release];							// Release the current sprite.
	SpriteImage = [OriginalSpriteImage copy];		// Get a copy of the original sprite template.
	
	[SpriteImage setSize:NSMakeSize( width, height )];		// Set the sprite's size.
	
	
	if ( center )
	{
		SpritePos.x -= ( ( [SpriteImage size].width - oldWidth ) / 2 );		// Center horizontally.
		SpritePos.y -= ( ( [SpriteImage size].height - oldHeight ) / 2 );	// Center vertically.
	}
	
	
	if ( width < oldWidth || height < oldHeight )
	{
		[self setDirtyRect:NSUnionRect( PreviousFrame, [self frame] )];	// Set the dirty rectangle.
	}
	else
	{
		[self setDirtyRect:[self frame]];
	}
}





/*****
		scaleByPercent:center:
		
		This method scales the sprite proportionately based on the passed in value.
		The passed in value must be from 1 to 100 (representing percentage).
*****/

- (void)scaleByPercent:(float)percent center:(BOOL)center
{
	float			percValue = percent / 100;
	float			hPixelValue = percValue * [OriginalSpriteImage size].width;
	float			vPixelValue = percValue * [OriginalSpriteImage size].height;


	[self newWidth:hPixelValue newHeight:vPixelValue center:center];		// Scale the sprite.
}





/*****
		scaleToPercentage:byPercInc:atRate:center:
		
		This method scales this sprite to the desired size using the passed-in increment
		and rate values. The sprite is scaled on a timer, allowing for a smooth scaling,
		animation effect.
*****/

- (void)scaleToPercentage:(float)scaleTo byPixelInc:(float)pixelInc atRate:(float)rate center:(BOOL)center
{
	if ( ScaleTimer != nil )		// If the scale timer is valid, then invalidate it.
	{
		[ScaleTimer invalidate];
		ScaleTimer = nil;
	}
	
	
	if ( scaleTo < 1.0 || scaleTo == 100.0 )
	{
		return;
	}
	
	
	
	float			currentWidth = [SpriteImage size].width;
	float			currentHeight = [SpriteImage size].height;
	NSDictionary	*timerInfoDict = [NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithFloat:scaleTo], @"scale to",
									 [NSNumber numberWithFloat:pixelInc], @"pixel inc",
									 [NSNumber numberWithFloat:currentWidth], @"width",
									 [NSNumber numberWithFloat:currentHeight], @"height",
									 [NSNumber numberWithBool:center], @"center", nil];
	
	
	ScaleTimer = [NSTimer scheduledTimerWithTimeInterval:rate target:self selector:@selector(scale:) userInfo:timerInfoDict repeats:YES];
}





/*****
		scaleToOriginalSizeByPixelInc:atRate:center:
		
		This method scales this sprite to it's original size (the size of the OriginalSpriteImage)
		by the passed-in pixel increment and desired rate.
*****/

- (void)scaleToOriginalSizeByPixelInc:(float)pixelInc atRate:(float)rate center:(BOOL)center
{
	NSSize				originalSize = [OriginalSpriteImage size];
	NSSize				currentSize = [self size];
	float				scaleTo = [GameUtilities percentageFromPixelValue:originalSize.width ofPixelsLength:currentSize.width];
	
	
	[self scaleToPercentage:scaleTo byPixelInc:pixelInc atRate:rate center:center];
}





/*****
		scale:
		
		This method scales the sprite's size each time it's called by the scale timer.
		When the sprite has reached the desired size, the scale timer is invalidated.
*****/

- (void)scale:(NSTimer *)timer
{
	NSDictionary	*timerInfoDict = [timer userInfo];
	NSSize			currentSize = [self size];
	float			currentWidth = currentSize.width;
	float			currentHeight = currentSize.height;
	float			originalWidth = [[timerInfoDict objectForKey:@"width"] floatValue];
	float			originalHeight = [[timerInfoDict objectForKey:@"height"] floatValue];
	float			pixelInc = [[timerInfoDict objectForKey:@"pixel inc"] floatValue];
	float			scaleTo = [[timerInfoDict objectForKey:@"scale to"] floatValue];
	float			desiredWidth = [GameUtilities pixelValueOfPercentage:scaleTo ofPixelLength:originalWidth];
	float			desiredHeight = [GameUtilities pixelValueOfPercentage:scaleTo ofPixelLength:originalHeight];
	BOOL			center = [[timerInfoDict objectForKey:@"center"] boolValue];
	

	
	// Scaler the sprite.
	
	if ( scaleTo > 100.0 )
	{
		[self newWidth:currentWidth + pixelInc newHeight:currentHeight + pixelInc center:NO];
	}
	else
	{
		[self newWidth:currentWidth - pixelInc newHeight:currentHeight - pixelInc center:NO];
	}
	

	currentSize = [self size];		// Get the sprite's new size.
	
	
	if ( center )	// Make the sprite maintain it's center while scaling?
	{
		SpritePos.x -= ( ( currentSize.width - currentWidth ) / 2 );
		SpritePos.y -= ( ( currentSize.height - currentHeight ) / 2 );
	}


	
	if ( scaleTo > 100.0 )		// Has the sprite reached the desired size yet?
	{
		if ( currentSize.width < desiredWidth || currentSize.height < desiredHeight )
		{
			return;
		}
	}
	else
	{
		if ( currentSize.width > desiredWidth || currentSize.height > desiredHeight )
		{
			return;
		}
	}
	
	
	[self newWidth:desiredWidth newHeight:desiredHeight center:NO];		// Make the sprite it's desired size. (It may not be exact after scaling.)
	
	[ScaleTimer invalidate];
	ScaleTimer = nil;
}



#pragma mark -
#pragma mark Rotation



/*****
		rotateByDegrees:
		
		This method rotates the sprite right (clockewise) or left (counterclockwise)
		on it's z axis.
*****/

- (void)rotateByDegrees:(float)deg
{
	NSImage					*newImage = [[NSImage alloc] init];
	NSImage					*tempImage = [OriginalSpriteImage copy];
	NSAffineTransform       *rotPath = [NSAffineTransform transform];
	NSAffineTransform		*rotImage = [NSAffineTransform transform];
	NSSize					newSize;
	NSPoint					imageOffset;
	NSRect					newRect = NSMakeRect( 0.0, 0.0, DesiredSize.width, DesiredSize.height );	// Create a new rectangle for the resized and rotated sprite image.
	NSBezierPath			*tempPath = [NSBezierPath bezierPathWithRect:newRect];						// Create the path to be rotated.


	[tempImage setSize:DesiredSize];		// Set the desired sprite size before rotating.

	
	
	// Rotate the Bezier Path. The purpose here is to rotate a path rectangle that is the same size
	// as the sprite image. This way we can get the path's bounds, and this will be the sprite's new
	// rectangle after rotation.

	[rotPath translateXBy:DesiredSize.width / 2 yBy:DesiredSize.height / 2];	// Place origin point at the center of the path rectangle.
	[rotPath rotateByDegrees:deg];												// Rotate the path around it's origin.

	[tempPath transformUsingAffineTransform:rotPath];		// Rotate the invisible path rectangle.

	newSize = [tempPath bounds].size;		// Get the size of the rotated path.
	[newImage setSize:newSize];				// Set the new sprite image to that new size.

	
	
	// Create a rotation for the sprite image, now that we know how big it's rectangle needs to be.
	
	[rotImage translateXBy:newSize.width / 2 yBy:newSize.height / 2];				// Place the sprite image's origin at the center of it's rectangle.
	[rotImage rotateByDegrees:deg];													// Rotate the sprite image around it's center.
	[rotImage translateXBy:-( newSize.width / 2 ) yBy:-( newSize.height / 2 )];		// Restore the sprite image's origin to it's previous position.
	
	
	
	// To keep the rotated image centered, and to prevent it from "running away", it's origin is
	// offset in the opposite direction that the rectangle is growing in.
	
	imageOffset = NSMakePoint( ( newSize.width - DesiredSize.width ) / 2, ( newSize.height - DesiredSize.height ) / 2 );		// Get the rotated image rectangle's offset.

	
	
	// Restore the sprite's origin to where it was before the previous rotation. (This prevents it from "running away".)
	
	SpritePos.x += SpriteOffsetPos.x;
	SpritePos.y += SpriteOffsetPos.y;


	
	// Copy the sprite image onto the new transformed image.

	[newImage lockFocus];		// Lock focus on the new image.

	[rotImage concat];		// Do transformations on the new image.

	[tempImage drawAtPoint:imageOffset fromRect:newRect operation:NSCompositeSourceOver fraction:1.0];		// Draw the sprite image onto it's new transformed image.

	[newImage unlockFocus];		// Unlock focus from the new image.



	[SpriteImage release];					// Free the previous sprite image.
	SpriteImage = [newImage retain];		// Retain a reference to the new image.

	[newImage release];			// Free the new image.
	[tempImage release];		// Free the temporary image.
	
	
	
	// After the rotation, the image has been resized. Reposition it's origin to keep it centered.
	
	SpritePos.x -= imageOffset.x;
	SpritePos.y -= imageOffset.y;

	
	SpriteOffsetPos = imageOffset;		// Save the offset so it can be restored the next time around.
}





/*****
		setSpinRate:degreesPerRotation:
		
		This method starts the sprite spinning. The passed in spin rate value is
		used as a timer value.
*****/

- (void)setSpinRate:(float)rate degreesPerRotation:(float)degPerRotation
{
	if ( SpinTimer != nil )		// If the timer is currently firing, then release it.
	{
		[SpinTimer invalidate];
		SpinTimer = nil;
	}
	
	
	DegreesToRotate = degPerRotation;
	
	if ( rate > 0.0 )		// If the rate is greater than zero, then start spinning.
	{
		SpinTimer = [NSTimer scheduledTimerWithTimeInterval:rate target:self selector:@selector(spin:) userInfo:nil repeats:YES];
	}
}





/*****
		spin rate = spinRate
		
		This method returns the rate at which this sprite is spinning.
*****/

- (float)spinRate
{
	if ( SpinTimer != nil )
	{
		return	[SpinTimer timeInterval];
	}
	
	
	return	0.0;
}





/*****
		spin:
		
		This method is called by the spin timer, making the sprite spin.
*****/

- (void)spin:(NSTimer *)timer
{
	Rotation += DegreesToRotate;	// Next angle...
	
	
	if ( Rotation > 360.0 || Rotation < -360.0 )	// Keep the angle between zero and positive or negative 360. 
	{
		Rotation = 0.0;
	}

	
	[self rotateByDegrees:Rotation];		// Rotate!
}



#pragma mark -
#pragma mark Shearing



/*****
		shearXby:yBy:
		
		This method creates a shearing effect on the sprite.
		
		NOTE: at present, shearing has issues where the sprite can get clipped,
			  hiding part or all of it.
*****/

- (void)shearXBy:(float)xShear yBy:(float)yShear
{
	NSAffineTransform			*shearTransform = [NSAffineTransform transform];
	NSAffineTransformStruct		transformStruct = [shearTransform transformStruct];
	NSImage						*newImage = [[NSImage alloc] init];
	NSImage						*tempImage = [OriginalSpriteImage copy];
	NSRect						oldSpritePosRect = [self frame];
	NSRect						newRect = NSMakeRect( 0.0, 0.0, DesiredSize.width, DesiredSize.height );
	NSPoint						imageOffset = NSZeroPoint;


	[tempImage setScalesWhenResized:YES];
	[tempImage setSize:DesiredSize];			// Set the desired sprite size before shearing.
	
	

	// Shear a bezier path and get it's new bounds.

	NSBezierPath		*tempPath = [NSBezierPath bezierPathWithRect:newRect];		// Create a new path.


	transformStruct.m21 = xShear;		// Set horizontal shear value.
	transformStruct.m12 = yShear;		// Set vertical shear value. 
	
	[shearTransform setTransformStruct:transformStruct];			// Set the new transformation structure.
	[tempPath transformUsingAffineTransform:shearTransform];		// Shear the path!

	
	NSSize				newSize = [tempPath bounds].size;		// Get the size of the sheared path.


	[newImage setSize:newSize];				// Set the new image to the same size as the path.
	
	
	if ( xShear < 0.0 )		// If shearing in a negatize direction, then calculate the image offset. (This is to keep it centered horizonmtally.)
	{
		imageOffset.x = newSize.width - DesiredSize.width;
	}


	if ( yShear < 0.0 )		// If shearing in a negatize direction, then calculate the image offset. (This is to keep it centered vertically.)
	{
		imageOffset.y = newSize.height - DesiredSize.height;
	}
	
	
	
	// Restore the sprite's origin to where it was before the previous shearing. (This prevents it from "running away".)
	
	SpritePos.x += SpriteOffsetPos.x;
	SpritePos.y += SpriteOffsetPos.y;



	// Copy the sprite image onto the new transformed image.
	
	[newImage lockFocus];		// Lock focus on the new image.
	[shearTransform concat];	// Apply the transformation to the new image.
	
	[tempImage drawAtPoint:imageOffset fromRect:newRect operation:NSCompositeSourceOver fraction:1.0];		// Draw the sprite onto it's new transformed image.
	
	[newImage unlockFocus];		// Unlock focus from the new image.
	
	
	[SpriteImage release];				// Free the previous sprite image.
	SpriteImage = [newImage retain];	// Retain a reference to the new image.
	
	[newImage release];			// Free the new image.
	[tempImage release];		// Free the temporary image.

	
	
	// After the shearing, the image has been resized. Reposition it's origin to keep it centered.
	
	SpritePos.x -= imageOffset.x;
	SpritePos.y -= imageOffset.y;
	
	
	SpriteOffsetPos = imageOffset;		// Save the offset so it can be restored the next time around.

	[self setDirtyRect:NSUnionRect( oldSpritePosRect, [self frame] )];		// Set the dirty rectangle.
}



#pragma mark -
#pragma mark Strobing



/*****
		strobeAtRate:
		
		This method makes the sprite strobe on and off at the given rate. The rate is
		the delay time between strobes. A rate of zero or any negative value stops the
		strobing effect.
*****/

- (void)strobeAtRate:(float)rate
{
	if ( StrobeTimer != nil )		// Check if the strobe timer is firing.
	{
		[StrobeTimer invalidate];	// If so, invalidate it.
		StrobeTimer = nil;
	}
	
	
	[self setOpacity:SavedOpacity];		// Make sure the sprite is visible.
	
	
	if ( rate > 0.0 )		// If the rate is greater than zero, then start strobing.
	{
		StrobeTimer = [NSTimer scheduledTimerWithTimeInterval:rate target:self selector:@selector(strobe:) userInfo:nil repeats:YES];
	}
}





/*****
		rate = strobeRate
		
		This method returns this sprite's rate of strobing.
*****/

- (float)strobeRate
{
	if ( StrobeTimer != nil )
	{
		return	[StrobeTimer timeInterval];
	}
	
	
	return	0.0;
}





/*****
		strobe:
		
		This method is invoked by the strobeTimer, making the sprite strobe by switching back
		and forth between an opacity of zero and it's current opacity level.
*****/

- (void)strobe:(NSTimer *)timer
{
	[self setVisible:!VisibleFlag];
}



#pragma mark -
#pragma mark Fading



/*****
		fadeTo:atRate:
		
		This method fades the sprite from it's current opacity level to the desired level.
*****/

- (void)fadeTo:(float)level atRate:(float)rate
{
	[self fadeFrom:Opacity fadeTo:level atRate:rate];
}





/*****
		fadeFrom:fadeTo:atRate:
		
		This method fades the sprite from one opacity level to another
		at the desired rate.
*****/

- (void)fadeFrom:(float)from fadeTo:(float)to atRate:(float)rate
{
	[self fadeFrom:from fadeTo:to inIncrementsOf:0.01 atRate:rate];
}





/*****
		fadeFrom:fadeTo:inIncrementsOf:atRate:
		
		This method fades the sprite from one opacity level to another at
		the desired increments and rate.
*****/

- (void)fadeFrom:(float)from fadeTo:(float)to inIncrementsOf:(float)fadeIncrement atRate:(float)rate
{
	if ( FadeTimer != nil )		// Invalidate the fading timer if it's currently active.
	{
		[FadeTimer invalidate];
		FadeTimer = nil;
	}

	
	if ( to != Opacity && rate > 0.0 )		// Set up the fading as long as it's not fading to it's current level and the fade rate is greater than zero.
	{
		FadeFromOpacity = from;
		FadeToOpacity = to;
		
		
		if ( FadeFromOpacity < 0.0 )		// Do not let the fade start point be less than zero.
		{
			FadeFromOpacity = 0.0;
		}
		
		
		if ( FadeFromOpacity > 1.0 )		// Do not let the fade start point be greater than one.
		{
			FadeFromOpacity = 1.0;
		}
		
		
		if ( FadeToOpacity < 0.0 )			// Do not let the fade destination point be less than zero.
		{
			FadeToOpacity = 0.0;
		}
		
		
		if ( FadeToOpacity > 1.0 )			// Do not let the fade destination point be greater than one.
		{
			FadeToOpacity = 1.0;
		}


		Opacity = FadeFromOpacity;

		FadeTimer = [NSTimer scheduledTimerWithTimeInterval:rate target:self selector:@selector(fading:)
						userInfo:[NSDictionary dictionaryWithObject:[NSNumber numberWithFloat:fadeIncrement] forKey:@"increment"] repeats:YES];
	}
}





/*****
		fading:
		
		This method is called by the fading timer.
*****/

- (void)fading:(NSTimer *)timer
{
	if ( VisibleFlag )		// Only fade when the sprite is visible.
	{
		float			fadeIncrement = [[[timer userInfo] objectForKey:@"increment"] floatValue];
		
		
		if ( FadeToOpacity > FadeFromOpacity )		// Fade in or out?
		{
			Opacity += fadeIncrement;		// Fade in.
			
			
			if ( Opacity > FadeToOpacity )		// If faded in all the way, then stop the fading timer.
			{
				Opacity = FadeToOpacity;
				
				[self stopFading];
			}
		}
		else
		{
			Opacity -= fadeIncrement;		// Fade out.
			
			
			if ( Opacity < FadeToOpacity )		// If faded out all the way, then stop the fading timer.
			{
				Opacity = FadeToOpacity;
				
				[self stopFading];
			}
		}
	}
}





/*****
		stopFading
		
		This method stops the sprite from fading, leaving it at it's current opacity level.
*****/

- (void)stopFading
{
	if ( FadeTimer != nil )		// If the fade timer is still active, then release it.
	{
		[FadeTimer invalidate];
		FadeTimer = nil;
	}
	
	
	SavedOpacity = Opacity;		// Save the sprite's current opacity level.
}



#pragma mark -
#pragma mark Mirror



/*****
		hMirror
		
		This method flips the sprite image horizontally.
*****/

- (void)hMirror
{
	[self flipSpriteHorizontally:YES];
}





/*****
		vMirror
		
		This method flips the sprite image vertically.
*****/

- (void)vMirror
{
	[self flipSpriteHorizontally:NO];
}





/*****
		flipSpriteHorizontally:
		
		This method flips the sprite horizontally if YES is passed in or vertically if NO is passed in.
*****/

- (void)flipSpriteHorizontally:(BOOL)horizontal
{
	NSAffineTransform			*mirrorTrans = [NSAffineTransform transform];
	NSSize						size = [SpriteImage size];
	NSImage						*newImage = [[NSImage alloc] initWithSize:size];

	
	
	if ( horizontal )		// Flip horizontally or vartically?
	{
		[mirrorTrans scaleXBy:-1.0 yBy:1.0];					// Flip horizontally...
		[mirrorTrans translateXBy:-size.width yBy:0.0];
	}
	else
	{
		[mirrorTrans scaleXBy:1.0 yBy:-1.0];					// Flip vertically...
		[mirrorTrans translateXBy:0.0 yBy:-size.height];
	}
	

	[newImage lockFocus];		// Lock focus on the new image.
	
	[mirrorTrans concat];		// Apply the mirror transformation on the new image.
	
	[SpriteImage drawAtPoint:NSZeroPoint fromRect:NSMakeRect( 0.0, 0.0, size.width, size.height ) operation:NSCompositeSourceOver fraction:1.0];

	[newImage unlockFocus];		// Unlock focus from the new image.
	
	
	[SpriteImage release];				// Release the current sprite image.
	[OriginalSpriteImage release];		// Release the original sprite image template.
	
	OriginalSpriteImage = [newImage copy];		// Copy the new mirrored image and make it the new sprite image template.
	SpriteImage = [newImage retain];			// Make the current sprite image match the new sprite image template.


	[newImage release];		// Release the new sprite image.
}



#pragma mark -
#pragma mark Shadow



/*****
		setShadowOffset:color:blur:
		
		This method sets up a shadow for this sprite.
*****/

- (void)setShadowOffset:(NSSize)offset color:(NSColor *)color blur:(float)blur
{
	// Setup this sprite's shadow.
	
	[SpriteShadow release];
	SpriteShadow = [[NSShadow alloc] init];
	
	[SpriteShadow setShadowOffset:offset];
	[SpriteShadow setShadowColor:color];
	[SpriteShadow setShadowBlurRadius:blur];
	
	
	
	// Setup the "no shadow" for unsetting this sprite's shadow after it has been drawn.
	
	[NoShadow release];
	NoShadow = [[NSShadow alloc] init];

	[NoShadow setShadowColor:[NSColor clearColor]];
}




/*****
		sprite's shadow = shadow
		
		This method returns a reference to this sprite's shadow.
*****/

- (NSShadow *)shadow
{
	return	SpriteShadow;
}





/*****
		removeShadow
		
		This method removes this sprite's shadow.
*****/

- (void)removeShadow
{
	[SpriteShadow release];
	SpriteShadow = nil;
	
	[NoShadow release];
	NoShadow = nil;
}



#pragma mark -
#pragma mark Draw Sprite



/*****
		setDirtyRect:
		
		This method sets the dirty rectangle to a slightly expanded passed-in rectangle.
*****/

- (void)setDirtyRect:(NSRect)dRect
{
	if ( SpriteShadow != nil )		// If this sprite has a shadow, then expand the dirty rectangle to include it.
	{
		NSSize		shadowOffset = [SpriteShadow shadowOffset];
		float		shadowXOffset = shadowOffset.width;
		float		shadowYOffset = shadowOffset.height;
		float		blurRadius = [SpriteShadow shadowBlurRadius];
		float		newXPos = dRect.origin.x;
		float		newYPos = dRect.origin.y;
		float		newWidth = dRect.size.width;
		float		newHeight = dRect.size.height;


		if ( shadowXOffset < 0.0 )
		{
			newXPos += shadowXOffset;
			newWidth += abs( (int)shadowXOffset );
		}
		
		if ( shadowYOffset < 0.0 )
		{
			newYPos += shadowYOffset;
			newHeight += abs( (int)shadowYOffset );
		}
		
		if ( shadowXOffset > 0.0 )
		{
			newWidth += shadowXOffset;
		}
		
		if ( shadowYOffset > 0.0 )
		{
			newHeight += shadowYOffset;
		}
		
		if ( blurRadius > 15.0 )
		{
			blurRadius = 15.0;
		}


		dRect = NSMakeRect( newXPos, newYPos, newWidth, newHeight );
		dRect = NSInsetRect( dRect, -blurRadius, -blurRadius );
	}
	
	
	DirtyRect = NSInsetRect( dRect, -2.0, -2.0 );
}





/*****
		dirty rectangle = dirtyRect
		
		This method returns the sprite's rectangle plus the dirty region around it. This
		is to prevent pieces of sprite from being left behind.
*****/

- (NSRect)dirtyRect
{
	return	DirtyRect;
}





/*****
			draw    
			
			This method draws this sprite.
*****/

- (void)draw
{
	if ( SpriteShadow != nil )		// Is there a shadow?
	{
		[SpriteShadow set];		// Set this sprite's shadow.
	}
	
	
	[SpriteImage drawAtPoint:SpritePos fromRect:[self bounds] operation:NSCompositeSourceOver fraction:Opacity];	// Draw this sprite.
	
	PreviousFrame = [self frame];

	
	if ( NoShadow != nil )		// If this sprite has a shadow, then unset it.
	{
		[NoShadow set];
	}
}



#pragma mark -
#pragma mark Boundary Detection



/*****
		setBoundaryLeft:top:right:bottom:autoConfineToBoundary:
		
		This method sets the sprite's boundary.
*****/

- (void)setBoundaryLeft:(float)left top:(float)top right:(float)right bottom:(float)bottom autoConfineToBoundary:(BOOL)confine
{
	LeftBoundary = left;
	TopBoundary = top;
	RightBoundary = right;
	BottomBoundary = bottom;
	
	AutoConfineToBoundaryFlag = confine;
}





/*****
		setBoundaryRect:autoConfineToBoundary:
		
		This method sets the sprite's boundary using the passed-in rectangle.
*****/

- (void)setBoundaryRect:(NSRect)bRect autoConfineToBoundary:(BOOL)confine
{
	LeftBoundary = bRect.origin.x;
	BottomBoundary = bRect.origin.y;
	RightBoundary = ( LeftBoundary + bRect.size.width ) - 1.0;
	TopBoundary = ( BottomBoundary + bRect.size.height ) - 1.0;
	
	AutoConfineToBoundaryFlag = confine;
}





/*****
		rectangle = boundaryRect
		
		This method returns this sprite's boundary rectangle.
*****/

- (NSRect)boundaryRect
{
	return	NSMakeRect( LeftBoundary, BottomBoundary, RightBoundary - LeftBoundary, TopBoundary - BottomBoundary );
}





/*****
		setAutoConfineToBoundary:
		
		This method sets whether this sprite is automatically confined
		within it's defined boundary.
*****/

- (void)setAutoConfineToBoundary:(BOOL)confine
{
	AutoConfineToBoundaryFlag = confine;
}





/*****
		yes/no = autoConfineToBoundary
		
		This method returns a boolean indicating whether this sprite is
		automatically confined within it's defined boundary.
*****/

- (BOOL)autoConfineToBoundary
{
	return	AutoConfineToBoundaryFlag;
}





/*****
		reached boundary = boundaryReached
		
		This method checks if this sprite has reached any of it's boundary edges. If it has, then
		a definition for that edge is returned. Otherwise, a definition for no edge is returned.
*****/

- (int)boundaryReached
{
	if ( SpritePos.x < LeftBoundary )
	{
		return	SPRITE_REACHED_LEFT_BOUNDARY;
	}
	
	
	if ( SpritePos.x + [SpriteImage size].width > RightBoundary )
	{
		return	SPRITE_REACHED_RIGHT_BOUNDARY;
	}
	
	
	if ( SpritePos.y + [SpriteImage size].height > TopBoundary )
	{
		return	SPRITE_REACHED_TOP_BOUNDARY;
	}
	
	
	if ( SpritePos.y < BottomBoundary )
	{
		return	SPRITE_REACHED_BOTTOM_BOUNDARY;
	}

	
	return	SPRITE_REACHED_NO_BOUNDARY;
}





/*****
		reached boundary = checkAndConfineToBoundary
		
		
		This method checks if this sprite has reached any of it's boundary edges. This is ideal
		for keeping it within a rectangular boundary. If this sprite has, then a definition for
		that edge is returned. Otherwise, a definition for no edge is returned.
*****/

- (int)checkAndConfineToBoundary
{
	int			aBoundary = [self boundaryReached];
	
	
	
	// Has this sprite reached the left edge of it's boundary?

	if ( aBoundary == SPRITE_REACHED_LEFT_BOUNDARY )
	{
		SpritePos.x = LeftBoundary;
		SpriteDir.x = -SpriteDir.x;
	}
	
	
	
	// Has this sprite reached the right edge of it's boundary?
	
	if ( aBoundary == SPRITE_REACHED_RIGHT_BOUNDARY )
	{
		SpritePos.x = RightBoundary - [SpriteImage size].width;
		SpriteDir.x = -SpriteDir.x;
	}

	
	
	// Has this sprite reached the top edge of it's boundary?
	
	if ( aBoundary == SPRITE_REACHED_TOP_BOUNDARY )
	{
		SpritePos.y = TopBoundary - [SpriteImage size].height;
		SpriteDir.y = -SpriteDir.y;
	}
	
	
	
	// Has this sprite reached the bottom edge of it's boundary?
	
	if ( aBoundary == SPRITE_REACHED_BOTTOM_BOUNDARY )
	{
		SpritePos.y = BottomBoundary;
		SpriteDir.y = -SpriteDir.y;
	}
	
	
	if ( aBoundary != SPRITE_REACHED_NO_BOUNDARY && ReachedBoundsTarget != nil )
	{
		[ReachedBoundsTarget performSelector:ReachedBoundsAction withObject:self];
	}


	return	aBoundary;
}



#pragma mark -
#pragma mark Collision Detection



/*****
		setCollisionAreaLeftOffset:topOffset:rightOffset:bottomOffset:
		
		This method sets the left, top, right, and bottom offsets for
		the collision rectangle.
*****/

- (void)setCollisionAreaLeftOffset:(float)leftOffset topOffset:(float)topOffset rightOffset:(float)rightOffset bottomOffset:(float)bottomOffset
{
	CollisionLeftOffset = leftOffset;
	CollisionTopOffset = topOffset;
	CollisionRightOffset = rightOffset;
	CollisionBottomOffset = bottomOffset;
}




/*****
		setCollisionAreaWidthOffset:heightOffset:
		
		This method sets the left/right and top/bottom distance between the
		sprite rectangle and the collision rectangle.
*****/

- (void)setCollisionAreaWidthOffset:(float)widthOffset heightOffset:(float)heightOffset
{
	[self setCollisionAreaLeftOffset:widthOffset topOffset:heightOffset rightOffset:widthOffset bottomOffset:heightOffset];
}





/*****
		setCollisionAreaOffset:
		
		This method sets the distance between the sprite rectangle's four edges
		and the collision rectangle's four edges.
*****/

- (void)setCollisionAreaOffset:(float)areaOffset
{
	[self setCollisionAreaLeftOffset:areaOffset topOffset:areaOffset rightOffset:areaOffset bottomOffset:areaOffset];
}





/*****
		rectangle = collisionArea
		
		This method returns this sprite's collision rectangle.
*****/

- (NSRect)collisionArea
{
	float		collisionXPos = SpritePos.x - CollisionLeftOffset;
	float		collisionYPos = SpritePos.y - CollisionBottomOffset;
	float		collisionWidth = CollisionLeftOffset + [self size].width + CollisionRightOffset;
	float		collisionHeight = CollisionBottomOffset + [self size].height + CollisionTopOffset;


	return	NSMakeRect( collisionXPos, collisionYPos, collisionWidth, collisionHeight );
}





/*****
		yes/no = collidedWithPoint:
		
		This method checks if the passed-in point is within this sprite's rectangle.
*****/

- (BOOL)collidedWithPoint:(NSPoint)thePoint
{
	if ( CollisionDetectionFlag && NSPointInRect( thePoint, [self collisionArea] ) )
	{
		if ( CollisionTarget != nil )
		{
			[CollisionTarget performSelector:CollisionAction withObject:self];
		}
		
		return	YES;		// The passed-in point is within this sprite's rectangle.
	}
	
	
	return	NO;		// The passed-in point is not within this sprite's rectangle.
}





/*****
		yes/no = collidedWithRect:
		
		This method checks if this sprite's rectangle overlaps the passed-in rectangle.
*****/

- (BOOL)collidedWithRect:(NSRect)theRect
{
	if ( CollisionDetectionFlag && NSIntersectsRect( theRect, [self collisionArea] ) )
	{
		if ( CollisionTarget != nil )
		{
			[CollisionTarget performSelector:CollisionAction withObject:self];
		}
		
		return	YES;		// The sprite rectangle overlaps the passed-in rectangle.
	}


	return	NO;		// No collision.
}





/*****
		yes/no = collidedWithSprite:
		
		This method checks if this sprite has collided with the passed-in sprite.
		
		NOTE: not yet implemented. For now, this method just calls the collidedWithRect: method.
*****/

- (BOOL)collidedWithSprite:(GameSprite *)otherSprite
{
	if ( [otherSprite canCollide] && [self collidedWithRect:[otherSprite collisionArea]] )		// Check if this sprite's rectangle overlaps with the passed-in sprite's rectangle.
	{
		// Check if there is any actual overlapping between the two sprite's...

		
		return	YES;		// A collision occured.
	}


	return	NO;		// No collision.
}



@end
