/*
		GameUtilities.m


		This class contains general game utilities.
		
		Created by Ian Howle on 3/12/2006.
		Public Domain.
*/




#import "GameUtilities.h"



@implementation GameUtilities



/*****
		pixel value = pixelValueOfPercentage:ofPixelLength
		
		This method returns the pixel value of the percentage of the passed in pixel length.
*****/

+ (float)pixelValueOfPercentage:(float)perc ofPixelLength:(float)pixelLength
{
	float		percValue = perc / 100.0;
	float		pixelValue = pixelLength * percValue;
	
	
	return	pixelValue;
}





/*****
		percentage value = percentageFromPixelValue:ofPixelsLength
		
		This method returns what the percentage value of the passed-in pixel is of pixel length.
*****/

+ (float)percentageFromPixelValue:(float)pixelValue ofPixelsLength:(float)pixelLength
{
	return	( pixelValue / pixelLength ) * 100;
}





/*****
		origin point = originOfRect:centeredInRect:
		
		This method returns the origin of the passed-in rect after
		it has been centered in the passed-in inRect. 
*****/

+ (NSPoint)originOfRect:(NSRect)rect centeredInRect:(NSRect)inRect
{
	float		centeredXOrigin = [GameUtilities xOriginOfRect:rect centeredVerticallyInRect:inRect];
	float		centeredYOrigin = [GameUtilities yOriginOfRect:rect centeredHorizontallyInRect:inRect];


	return	NSMakePoint( centeredXOrigin, centeredYOrigin );
}




+ (float)xOriginOfRect:(NSRect)rect centeredVerticallyInRect:(NSRect)inRect
{
	float		rectWidth = rect.size.width;
	float		inRectWidth = inRect.size.width;
	float		inRectX = inRect.origin.x;


	return	( ( inRectWidth - rectWidth ) / 2 ) + inRectX;
}



+ (float)yOriginOfRect:(NSRect)rect centeredHorizontallyInRect:(NSRect)inRect
{
	float		rectHeight = rect.size.height;
	float		inRectHeight = inRect.size.height;
	float		inRectY = inRect.origin.y;


	return	( ( inRectHeight - rectHeight ) / 2 ) + inRectY;
}





/*****
		randomPointWithMinXValue:maxXValue:minYValue:maxYValue:
		
		This method returns a random point based on the passed-in minValue and minValue values.
*****/

+ (NSPoint)randomPointWithMinXValue:(float)minX maxXValue:(float)maxX minYValue:(float)minY maxYValue:(float)maxY
{
	int			randX = ( random() % (int)maxX ) + (int)minX;
	int			randY = ( random() % (int)maxY ) + (int)minY;


	return	NSMakePoint( (float)randX, (float)randY );		// Return the random point structure.
}





+ (float)randomDirWithMaxValue:(float)maxValue allowZero:(BOOL)allowZero
{
	float		randomDir;


	do
	{
		randomDir = ( random() % ( (int)maxValue + 2 ) ) - maxValue;
	
	}	while ( !allowZero && randomDir == 0 );


	return	randomDir;
}





+ (NSPoint)randomDirsWithMaxHorizontalValue:(float)hVAlue maxVerticalValue:(float)vValue allowZero:(BOOL)allowZero
{
	float		randomHDir = [GameUtilities randomDirWithMaxValue:hVAlue allowZero:allowZero];
	float		randomVDir = [GameUtilities randomDirWithMaxValue:vValue allowZero:allowZero];


	return	NSMakePoint( randomHDir, randomVDir );
}





/*****
		new master image = newMasterImageFromMasterImage:leftFrames:
		
		This method extracts the frames from the passed-in master image.
		The frames should be all left-facing. A new master image is created
		that is wide enough to accomodate all of the frames. The left frames
		and flipped versions of them that face right are added to the new
		master image, which is returned.
*****/

+ (NSImage *)newMasterImageFromMasterImage:(NSImage *)masterImage leftFrames:(NSMutableArray *)leftFrames
{
	NSSize			masterImageSize = [masterImage size];			// Get the size of the passed-in mater image.
	float			leftFramesWidth = 0.0;							// This will be the combined width of the passed-in left frames.
	int				newMasterImageWidth = masterImageSize.width;	// Add the passed-in master image width to the new master image width.
	int				numberOfLeftFrames = [leftFrames count];		// Get the total number of passed-in left frames.
	int				i;
	
	
	
	// Determine the combined length of all the passed-in left frames.

	for ( i = 0; i < numberOfLeftFrames; i++ )
	{
		NSRect			leftFrame = [[leftFrames objectAtIndex:i] rectValue];
		float			leftFrameWidth = leftFrame.size.width;

		
		leftFramesWidth += leftFrameWidth;
	}
	
	newMasterImageWidth += leftFramesWidth;		// Add the combined left frames width to the current width (this is to allow space for the flipped images).
	
	
	
	// Create the new master image.
	
	NSImage			*newMasterImage = [[[NSImage alloc] initWithSize:NSMakeSize( newMasterImageWidth, masterImageSize.height )] autorelease];
	
	
	
	// Copy the frames from the master image into the new master image.
	// (Place at the left half of the new master image.)
	
	[newMasterImage lockFocus];
	
	[masterImage drawAtPoint:NSMakePoint( 0.0, 0.0 ) fromRect:NSMakeRect( 0.0, 0.0, masterImageSize.width, masterImageSize.height )
		operation:NSCompositeSourceOver fraction:1.0];
	
	[newMasterImage unlockFocus];
	
	
	
	// Flip each of the master image's frames from left to right and add them to
	// the right of the new master image.
	
	float			firstFrameXPos;
	

	for ( i = 0; i < numberOfLeftFrames; i++ )
	{
		NSAffineTransform		*mirrorTrans = [NSAffineTransform transform];
		NSRect					leftFrame = [[leftFrames objectAtIndex:i] rectValue];
		NSImage					*flippedImage = [[NSImage alloc] initWithSize:leftFrame.size];
		
		
		if ( i == 0 )
		{
			firstFrameXPos = leftFrame.origin.x;
		}
		
		
		
		// Setup a verticle flip transformation.
		
		[mirrorTrans scaleXBy:-1.0 yBy:1.0];
		[mirrorTrans translateXBy:-leftFrame.size.width yBy:0.0];


		[flippedImage lockFocus];
		
		[mirrorTrans concat];
		
		[masterImage drawInRect:NSMakeRect( 0.0, 0.0, leftFrame.size.width, leftFrame.size.height )
			fromRect:leftFrame operation:NSCompositeSourceOver fraction:1.0];
		
		[flippedImage unlockFocus];
		
		
		
		// Copy the vertically flipped frame to the new master image.
		// (Add the flipped frames after the existing frames from left to right.)

		[newMasterImage lockFocus];
		
		[flippedImage drawAtPoint:NSMakePoint( firstFrameXPos + leftFramesWidth, 0.0 )
			fromRect:NSMakeRect( 0.0, 0.0, leftFrame.size.width, leftFrame.size.height ) operation:NSCompositeSourceOver fraction:1.0];

		[newMasterImage unlockFocus];
		
		
		
		// Add the new right frame to the frames array.

		NSValue			*rightFrame = [NSValue valueWithRect:NSMakeRect( firstFrameXPos + leftFramesWidth, 0.0, leftFrame.size.width, leftFrame.size.height )];

		[leftFrames addObject:rightFrame];
		
		
		leftFramesWidth += leftFrame.size.width;
	}


	return	newMasterImage;
}





/*****
		this game's directory path = gameDirectory
		
		This method returns the directory that this game file resides in.
*****/

+ (NSString *)gameDirectory
{
	NSString		*gamePath = [GameUtilities gamePath];
	NSString		*gameDir = [gamePath stringByDeletingLastPathComponent];


	return	gameDir;
}





/*****
		this game's path = gamePath
		
		This method returns a path to this game.
*****/

+ (NSString *)gamePath
{
	return	[[NSBundle mainBundle] bundlePath];
}



@end
