/*
		MyView.m


		This class represents the view where the sprites will be displayed. This class also
		serves as the sprite controller.
		
		Created by Ian Howle on 2/21/2005.
		Public Domain.
*/




#import "MyView.h"



@implementation MyView



- (id)initWithFrame:(NSRect)frameRect
{
	if ( ( self = [super initWithFrame:frameRect] ) != nil )
	{
		SpriteList = [[NSMutableArray alloc] init];		// This array holds all sprites controlled by this controller.
	
		BackgroundImage = [NSImage imageNamed:@"nightlivingroom"];
		SpriteTimer = nil;

		CollisionFlag = NO;
	}
	
	
	return self;
}





- (void)dealloc
{
	if ( SpriteTimer != nil )
	{
		[SpriteTimer invalidate];
	}
	
	[BackgroundImage release];
	[SpriteList release];

	[super dealloc];
}





- (void)drawRect:(NSRect)rect
{
	GameSprite			*theSprite;
	int					i;
	[BackgroundImage drawInRect:[self bounds] fromRect:NSMakeRect( 0.0, 0.0, [BackgroundImage size].width, [BackgroundImage size].height ) operation:NSCompositeSourceOver fraction:1.0];
	
	for ( i = 0; i < [SpriteList count]; i++ )		// Go through each sprite in the list.
	{
		theSprite = [SpriteList objectAtIndex:i];	// Get a sprite.
		
		[theSprite draw];
	}
}





/*****
		startSprites
		
		This method starts the sprite controller.
*****/

- (void)startSprites
{
	if ( SpriteTimer == nil || ![SpriteTimer isValid] )
	{
		SpriteTimer = [NSTimer scheduledTimerWithTimeInterval:0.01 target:self selector:@selector(runSprites:) userInfo:nil repeats:YES];
	}
}





/*****
		stopSprites
		
		This method stops the sprite controller.
*****/

- (void)stopSprites
{
	if ( SpriteTimer != nil )
	{
		[SpriteTimer invalidate];
		SpriteTimer = nil;
	}
}





/*****
		yes/no = areSpritesActive

		This method returns YES if all sprites are active, and NO if they are all inactive.
*****/

- (BOOL)areSpritesActive
{
	if ( SpriteTimer != nil )
	{
		return	[SpriteTimer isValid];
	}
	
	
	return	NO;
}





/*****
		success = addSprite:
		
		This method adds a new sprite to the existing sprite list.
*****/

- (BOOL)addSprite:(GameSprite *)theSprite
{
	NSString		*name = [theSprite identifier];
	int				i;


	if ( theSprite != nil )		// Make sure the sprite exists.
	{
		for ( i = 0; i < [SpriteList count]; i++ )		// Make sure no other sprites in the list have the same indentifier as this sprite.
		{
			if ( [[[SpriteList objectAtIndex:i] identifier] isEqualToString:name] )		// Check if a sprite with this name already exists.
			{
				NSLog( @"A sprite named %@ already exists.", name );
				
				return	NO;		// Sprite identifier name already exists.
			}
		}
		
		[SpriteList addObject:theSprite];	// Add the sprite to the sprite list.
		
		
		return	YES;	// Sprite successfully added.
	}
	
	
	NSLog( @"Sprite is NULL or name is blank." );


	return	NO;		// Passed in sprite object was NULL or no name was passed in.
}





/*****
		removeSprite:
		
		This method removes a sprite from the sprite list.
*****/

- (void)removeSprite:(NSString *)name
{
	int			i;
	

	for ( i = 0; i < [SpriteList count]; i++ )		// Search for the sprite to remove.
	{
		if ( [[[SpriteList objectAtIndex:i] identifier] isEqualToString:name] )
		{
			[SpriteList removeObjectAtIndex:i];		// Remove the sprite.
		}
	}
}





/*****
		sprite = spriteRef:
		
		This method returns a reference to a sprite.
*****/

- (GameSprite *)spriteRef:(NSString *)name
{
	int			i;
	
	
	for ( i = 0; i < [SpriteList count]; i++ )		// Find the desired sprite.
	{
		if ( [[[SpriteList objectAtIndex:i] identifier] isEqualToString:name] )
		{
			return	[SpriteList objectAtIndex:i];
		}
	}
	
	
	return	nil;
}





/*****
		runSprites:
		
		This method does sprite action for all sprites in the list.
*****/

- (void)runSprites:(NSTimer *)timer
{
	int			i;



	for ( i = 0; i < [SpriteList count]; i++ )		// Scan through each sprite.
	{
		GameSprite		*theSprite = [SpriteList objectAtIndex:i];	// Get a sprite from the list.


		if ( [theSprite spriteImage] != nil )		// Check if sprite needs to be drawn.
		{
			[theSprite move];
			[self setNeedsDisplayInRect:[theSprite dirtyRect]];
		}
		
		if ( CollisionFlag )
		{
			[self checkSpriteCollision:theSprite];		// Check if this sprite has collided with any others.
		}
	}
}





/*****
		checkSpriteCollision:
		
		This method checks if a sprite has collided with any other sprites.
*****/

- (void)checkSpriteCollision:(GameSprite *)theSprite
{
	GameSprite		*otherSprite;
	NSPoint			firstSpriteDir;
	NSPoint			secondSpriteDir;
	NSPoint			firstSpritePos;
	NSPoint			secondSpritePos;
	int				i;
	
	
	
	for ( i = 0; i < [SpriteList count]; i++ )
	{
		otherSprite = [SpriteList objectAtIndex:i];		// Get a reference to the other sprite.
		
		
		if ( ![[theSprite identifier] isEqualToString:[otherSprite identifier]] && [theSprite collidedWithSprite:otherSprite] )		// Check if the sprite has collided with another sprite.
		{
			firstSpriteDir = [theSprite spriteDir];				// Get the first sprite's direction.
			secondSpriteDir = [otherSprite spriteDir];			// Get the second sprite's direction.
			
			firstSpritePos = [theSprite origin];				// Get the first sprite's current position.
			secondSpritePos = [otherSprite origin];				// Get the second sprite's current position.
			
			
			firstSpriteDir.x = -firstSpriteDir.x;		// Reverse the first sprite's direction.
			firstSpriteDir.y = -firstSpriteDir.y;		// ...
			
			[theSprite setDirX:firstSpriteDir.x dirY:firstSpriteDir.y];
			[theSprite setPosX:firstSpritePos.x + firstSpriteDir.x posY:firstSpritePos.y + firstSpriteDir.y];
			
			
			secondSpriteDir.x = -secondSpriteDir.x;		// Reverse the second sprite's direction.
			secondSpriteDir.y = -secondSpriteDir.y;		// ...
			
			[otherSprite setDirX:secondSpriteDir.x dirY:secondSpriteDir.y];
			[otherSprite setPosX:secondSpritePos.x + secondSpriteDir.x posY:secondSpritePos.y + secondSpriteDir.y];
		}
	}
}





/*****
		toggleSpriteCollision:
		
		This method turns the sprite detection on or off.
*****/

- (void)setSpriteCollision:(BOOL)onOff
{
	CollisionFlag = onOff;
}





/*****
		relayer:
		
		This method reorders the sprite so that it is in a different layer (Z-order) than before.
*****/

- (IBAction)relayer:(id)sender
{


}



@end
